# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import jsonschema

from subiquitycore.context import with_context
from subiquitycore.controller import (
    BaseController,
    RepeatedController,
    Skip,
    )

log = logging.getLogger("subiquity.controller")


class SubiquityController(BaseController):

    autoinstall_key = None
    autoinstall_schema = None
    autoinstall_default = None

    def __init__(self, app):
        super().__init__(app)
        self.autoinstall_applied = False
        self.context.set('controller', self)
        self.setup_autoinstall()

    def setup_autoinstall(self):
        if self.app.autoinstall_config:
            with self.context.child("load_autoinstall_data"):
                ai_data = self.app.autoinstall_config.get(
                    self.autoinstall_key,
                    self.autoinstall_default)
                if ai_data is not None and self.autoinstall_schema is not None:
                    jsonschema.validate(ai_data, self.autoinstall_schema)
                self.load_autoinstall_data(ai_data)

    def load_autoinstall_data(self, data):
        """Load autoinstall data.

        This is called if there is an autoinstall happening. This
        controller may not have any data, and this controller may still
        be interactive.
        """
        pass

    @with_context()
    async def apply_autoinstall_config(self, context):
        """Apply autoinstall configuration.

        This is only called for a non-interactive controller. It should
        block until the configuration has been applied. (self.configured()
        is called after this is done).
        """
        pass

    def interactive(self):
        if not self.app.autoinstall_config:
            return True
        i_sections = self.app.autoinstall_config.get(
            'interactive-sections', [])
        if '*' in i_sections or self.autoinstall_key in i_sections:
            return True
        return False

    def configured(self):
        """Let the world know that this controller's model is now configured.
        """
        if self.model_name is not None:
            self.app.base_model.configured(self.model_name)

    def deserialize(self, state):
        pass

    def make_autoinstall(self):
        return {}


class NoUIController(SubiquityController):

    def start_ui(self):
        raise Skip

    def cancel(self):
        pass

    def interactive(self):
        return False


class RepeatedController(RepeatedController):

    autoinstall_key = None
    autoinstall_schema = None

    def __init__(self, orig, index):
        super().__init__(orig, index)
        self.autoinstall_applied = False

    async def apply_autoinstall_config(self):
        await self.orig.apply_autoinstall_config(index=self.index)

    def configured(self):
        self.orig.configured()

    def interactive(self):
        return self.orig.interactive()

    def make_autoinstall(self):
        return {}



# Copyright 2020 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from subiquity.controller import NoUIController


class PackageController(NoUIController):

    model_name = autoinstall_key = "packages"
    autoinstall_default = []
    autoinstall_schema = {
        'type': 'array',
        'items': {'type': 'string'},
        }

    def load_autoinstall_data(self, data):
        self.model[:] = data

    def make_autoinstall(self):
        return self.model



# Copyright 2020 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fcntl
import logging

from subiquitycore.async_helpers import run_in_thread

log = logging.getLogger('subiquity.lockfile')


class _LockContext:

    def __init__(self, lockfile, flags):
        self.lockfile = lockfile
        self.flags = flags
        self._kind = "???"
        if flags & fcntl.LOCK_EX:
            self._kind = "exclusive"
        elif flags & fcntl.LOCK_SH:
            self._kind = "shared"

    def __enter__(self):
        log.debug("locking %s %s", self._kind, self.lockfile.path)
        fcntl.flock(self.lockfile.fp, self.flags)
        return self

    async def __aenter__(self):
        return await run_in_thread(self.__enter__)

    def __exit__(self, etype, evalue, etb):
        log.debug("unlocking %s %s", self._kind, self.lockfile.path)
        fcntl.flock(self.lockfile.fp, fcntl.LOCK_UN)

    async def __aexit__(self, etype, evalue, etb):
        self.__exit__(etype, evalue, etb)


class Lockfile:

    def __init__(self, path):
        self.path = path
        self.fp = open(path, 'a+')

    def read_content(self):
        self.fp.seek(0)
        return self.fp.read()

    def write_content(self, content):
        self.fp.seek(0)
        self.fp.truncate()
        self.fp.write(content)
        self.fp.flush()

    def exclusive(self):
        return _LockContext(self, fcntl.LOCK_EX)

    def shared(self):
        return _LockContext(self, fcntl.LOCK_SH)

    def is_exclusively_locked(self):
        try:
            fcntl.flock(self.fp, fcntl.LOCK_SH | fcntl.LOCK_NB)
        except OSError:
            return True
        else:
            fcntl.flock(self.fp, fcntl.LOCK_UN)
            return False



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from subiquitycore.ui.frame import SubiquityCoreUI
from subiquitycore.view import BaseView

from subiquity.ui.views.help import HelpButton


log = logging.getLogger('subiquity.ui.frame')


class SubiquityUI(SubiquityCoreUI):

    block_input = False

    def __init__(self, app):
        self.app = app
        self.right_icon = HelpButton(app)
        super().__init__()

    def keypress(self, size, key):
        if not self.block_input:
            return super().keypress(size, key)

    def set_body(self, widget):
        super().set_body(widget)
        if isinstance(widget, BaseView):
            for overlay in self.app.global_overlays:
                widget.show_stretchy_overlay(overlay)



#!/usr/bin/env python3
# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from console_conf.controllers.identity import write_login_details_standalone
from subiquitycore.log import setup_logger


def main():
    setup_logger(dir='/var/log/console-conf')
    write_login_details_standalone()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from subiquitycore.models.network import NetworkModel

log = logging.getLogger('subiquity.models.network')


class NetworkModel(NetworkModel):

    def __init__(self):
        super().__init__("subiquity", support_wlan=False)
        self.override_config = None

    def render_config(self):
        if self.override_config is not None:
            return self.override_config
        else:
            return super().render_config()

    def render(self):
        return {
            'write_files': {
                'etc_netplan_installer': {
                    'path': 'etc/netplan/00-installer-config.yaml',
                    'content': self.stringify_config(self.render_config()),
                    },
                'nonet': {
                    'path': ('etc/cloud/cloud.cfg.d/'
                             'subiquity-disable-cloudinit-networking.cfg'),
                    'content': 'network: {config: disabled}\n',
                    },
                },
            }



# Copyright 2020 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from systemd import journal


def journald_listener(identifiers, callback, seek=False):
    reader = journal.Reader()
    args = []
    for identifier in identifiers:
        args.append("SYSLOG_IDENTIFIER={}".format(identifier))
    reader.add_match(*args)

    if seek:
        reader.seek_tail()

    def watch():
        if reader.process() != journal.APPEND:
            return
        for event in reader:
            callback(event)
    return reader.fileno(), watch



# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .filesystem import (
    FilesystemView,
    GuidedDiskSelectionView,
    )
from .identity import IdentityView
from .installprogress import ProgressView
from .keyboard import KeyboardView
from .welcome import WelcomeView
from .zdev import ZdevView
__all__ = [
    'FilesystemView',
    'GuidedDiskSelectionView',
    'IdentityView',
    'KeyboardView',
    'ProgressView',
    'WelcomeView',
    'ZdevView',
]



# Copyright 2020 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import platform
import subprocess

from subiquitycore.async_helpers import schedule_task
from subiquitycore.context import with_context
from subiquitycore.utils import arun_command, run_command

from subiquity.controller import SubiquityController

log = logging.getLogger("subiquity.controllers.restart")


class RebootController(SubiquityController):

    def __init__(self, app):
        super().__init__(app)
        self.context.set('hidden', True)

    def interactive(self):
        return self.app.interactive()

    @with_context()
    async def copy_logs_to_target(self, context):
        if self.opts.dry_run and 'copy-logs-fail' in self.app.debug_flags:
            raise PermissionError()
        target_logs = os.path.join(
            self.app.base_model.target, 'var/log/installer')
        if self.opts.dry_run:
            os.makedirs(target_logs, exist_ok=True)
        else:
            await arun_command(
                ['cp', '-aT', '/var/log/installer', target_logs])
        journal_txt = os.path.join(target_logs, 'installer-journal.txt')
        try:
            with open(journal_txt, 'w') as output:
                await arun_command(
                    ['journalctl', '-b'],
                    stdout=output, stderr=subprocess.STDOUT)
        except Exception:
            log.exception("saving journal failed")

    def reboot(self):
        if self.opts.dry_run:
            self.app.exit()
        else:
            if platform.machine() == 's390x':
                run_command(["chreipl", "/target/boot"])
            run_command(["/sbin/reboot"])

    @with_context()
    async def apply_autoinstall_config(self, context):
        await self.copy_logs_to_target(context=context)
        self.reboot()

    async def _run(self):
        await self.copy_logs_to_target()
        await self.app.controllers.InstallProgress.reboot_clicked.wait()
        self.reboot()

    def start_ui(self):
        schedule_task(self._run())

    def cancel(self):
        pass



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging

from subiquitycore.async_helpers import schedule_task
from subiquitycore.context import with_context
from subiquitycore.controllers.network import NetworkController

from subiquity.controller import SubiquityController
from subiquity.controllers.error import ErrorReportKind


log = logging.getLogger("subiquity.controllers.network")

MATCH = {
    'type': 'object',
    'properties': {
        'name': {'type': 'string'},
        'macaddress': {'type': 'string'},
        'driver': {'type': 'string'},
        },
    'additionalProperties': False,
    }

NETPLAN_SCHEMA = {
    'type': 'object',
    'properties': {
        'version': {
            'type': 'integer',
            'minimum': 2,
            'maximum': 2,
            },
        'ethernets': {
            'type': 'object',
            'properties': {
                'match': MATCH,
                }
            },
        'wifis': {
            'type': 'object',
            'properties': {
                'match': MATCH,
                }
            },
        'bridges': {'type': 'object'},
        'bonds': {'type': 'object'},
        'tunnels': {'type': 'object'},
        'vlans': {'type': 'object'},
        },
    'required': ['version'],
    }


class NetworkController(NetworkController, SubiquityController):

    ai_data = None
    autoinstall_key = "network"
    autoinstall_schema = {
        'oneOf': [
            NETPLAN_SCHEMA,
            {
                'type': 'object',
                'properties': {
                    'network': NETPLAN_SCHEMA,
                    },
                'required': ['network'],
            },
            ],
        }

    def __init__(self, app):
        super().__init__(app)
        app.note_file_for_apport("NetplanConfig", self.netplan_path)

    def load_autoinstall_data(self, data):
        if data is not None:
            self.ai_data = data
            # The version included with 20.04 accidentally required
            # that you put:
            #
            # network:
            #   network:
            #     version: 2
            #
            # in your autoinstall config. Continue to support that for
            # backwards compatibility.
            if 'network' in self.ai_data:
                self.ai_data = self.ai_data['network']

    def start(self):
        if self.ai_data is not None:
            self.model.override_config = {'network': self.ai_data}
            self.apply_config()
            if self.interactive():
                # If interactive, we want edits in the UI to override
                # the provided config. If not, we just splat the
                # autoinstall config onto the target system.
                schedule_task(self.unset_override_config())
        elif not self.interactive():
            self.initial_config = schedule_task(self.wait_for_initial_config())
        super().start()

    async def unset_override_config(self):
        await self.apply_config_task.wait()
        self.model.override_config = None

    @with_context()
    async def wait_for_initial_config(self, context):
        # In interactive mode, we disable all nics that haven't got an
        # address by the time we get to the network screen. But in
        # non-interactive mode we might get to that screen much faster
        # so we wait for up to 10 seconds for any device configured
        # to use dhcp to get an address.
        dhcp_events = set()
        for dev in self.model.get_all_netdevs(include_deleted=True):
            dev.dhcp_events = {}
            for v in 4, 6:
                if dev.dhcp_enabled(v) and not dev.dhcp_addresses()[v]:
                    dev.dhcp_events[v] = e = asyncio.Event()
                    dhcp_events.add(e)
        if not dhcp_events:
            return

        with context.child("wait_dhcp"):
            try:
                await asyncio.wait_for(
                    asyncio.wait({e.wait() for e in dhcp_events}),
                    10)
            except asyncio.TimeoutError:
                pass

    @with_context()
    async def apply_autoinstall_config(self, context):
        if self.ai_data is None:
            with context.child("wait_initial_config"):
                await self.initial_config
            self.update_initial_configs()
            self.apply_config(context)
        with context.child("wait_for_apply"):
            await self.apply_config_task.wait()
        self.model.has_network = bool(
            self.network_event_receiver.default_routes)

    async def _apply_config(self, *, context=None, silent=False):
        try:
            await super()._apply_config(context=context, silent=silent)
        except asyncio.CancelledError:
            # asyncio.CancelledError is a subclass of Exception in
            # Python 3.6 (sadface)
            raise
        except Exception:
            log.exception("_apply_config failed")
            self.model.has_network = False
            self.app.make_apport_report(
                ErrorReportKind.NETWORK_FAIL, "applying network",
                interrupt=True)
            if not self.interactive():
                raise

    def done(self):
        self.configured()
        super().done()

    def make_autoinstall(self):
        return self.model.render_config()['network']



# Copyright 2020 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urwid

from subiquitycore.tests import SubiTestCase
from subiquitycore.view import BaseView
from subiquitycore.ui.stretchy import Stretchy, StretchyOverlay
from subiquitycore.ui.utils import undisabled


class TestStretchy(Stretchy):
    was_opened = was_closed = False

    def __init__(self):
        super().__init__("", [urwid.Text("")], 0, 0)

    def opened(self):
        self.was_opened = True

    def closed(self):
        self.was_closed = True


class TestBaseView(SubiTestCase):

    def get_stretchy_chain(self, view):
        view = view._w
        r = []
        while isinstance(view, StretchyOverlay):
            r.append(view.stretchy)
            view = undisabled(view.bottom_w)
        return r

    def test_basic(self):
        bv = BaseView(urwid.Text(""))
        s = TestStretchy()
        self.assertEqual(self.get_stretchy_chain(bv), [])
        bv.show_stretchy_overlay(s)
        self.assertTrue(s.was_opened)
        self.assertEqual(self.get_stretchy_chain(bv), [s])

        bv.remove_overlay()
        self.assertTrue(s.was_closed)
        self.assertEqual(self.get_stretchy_chain(bv), [])

        bv.show_stretchy_overlay(s)
        bv.remove_overlay(s)
        self.assertEqual(self.get_stretchy_chain(bv), [])

    def make_view_with_overlays(self):
        bv = BaseView(urwid.Text(""))
        a = TestStretchy()
        b = TestStretchy()
        c = TestStretchy()
        bv.show_stretchy_overlay(a)
        bv.show_stretchy_overlay(b)
        bv.show_stretchy_overlay(c)
        return bv, a, b, c

    def test_multiple(self):
        bv, a, b, c = self.make_view_with_overlays()
        self.assertEqual(self.get_stretchy_chain(bv), [c, b, a])

        bv, a, b, c = self.make_view_with_overlays()
        bv.remove_overlay(a)
        self.assertTrue(a.was_closed)
        self.assertEqual(self.get_stretchy_chain(bv), [c, b])

        bv, a, b, c = self.make_view_with_overlays()
        bv.remove_overlay(b)
        self.assertTrue(b.was_closed)
        self.assertEqual(self.get_stretchy_chain(bv), [c, a])

        bv, a, b, c = self.make_view_with_overlays()
        bv.remove_overlay(c)
        self.assertTrue(c.was_closed)
        self.assertEqual(self.get_stretchy_chain(bv), [b, a])

        bv, a, b, c = self.make_view_with_overlays()
        bv.remove_overlay()
        self.assertTrue(c.was_closed)
        self.assertEqual(self.get_stretchy_chain(bv), [b, a])



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from subiquitycore.context import Context
from subiquity.controllers.filesystem import (
    FilesystemController,
    )
from subiquity.models.tests.test_filesystem import (
    make_disk,
    make_model,
    )
from subiquity.models.filesystem import (
    Bootloader,
    DeviceAction,
    )


class Thing:
    # Just something to hang attributes off
    pass


class MiniApplication:
    ui = signal = loop = None
    project = "mini"
    autoinstall_config = {}
    answers = {}
    opts = Thing()
    opts.dry_run = True
    opts.bootloader = None
    def report_start_event(*args): pass
    def report_finish_event(*args): pass


def make_controller(bootloader=None):
    app = MiniApplication()
    app.base_model = bm = Thing()
    app.context = Context.new(app)
    bm.target = '/target'
    bm.filesystem = make_model(bootloader)
    controller = FilesystemController(app)
    return controller


def make_controller_and_disk(bootloader=None):
    controller = make_controller(bootloader)
    return controller, make_disk(controller.model)


class TestFilesystemController(unittest.TestCase):

    def test_delete_encrypted_vg(self):
        controller, disk = make_controller_and_disk()
        spec = {
            'password': 'passw0rd',
            'devices': {disk},
            'name': 'vg0',
            }
        vg = controller.create_volgroup(spec)
        controller.delete_volgroup(vg)
        dm_crypts = [
            a for a in controller.model._actions if a.type == 'dm_crypt']
        self.assertEqual(dm_crypts, [])

    def test_can_only_add_boot_once(self):
        # This is really testing model code but it's much easier to test with a
        # controller around.
        for bl in Bootloader:
            controller, disk = make_controller_and_disk(bl)
            if DeviceAction.TOGGLE_BOOT not in disk.supported_actions:
                continue
            controller.add_boot_disk(disk)
            self.assertFalse(
                disk._can_TOGGLE_BOOT,
                "add_boot_disk(disk) did not make _can_TOGGLE_BOOT false "
                "with bootloader {}".format(bl))

    def assertIsMountedAtBootEFI(self, device):
        efi_mnts = device._m._all(type="mount", path="/boot/efi")
        self.assertEqual(len(efi_mnts), 1)
        self.assertEqual(efi_mnts[0].device.volume, device)

    def assertNotMounted(self, device):
        if device.fs():
            self.assertIs(device.fs().mount(), None)

    def add_existing_boot_partition(self, controller, disk):
        if controller.model.bootloader == Bootloader.BIOS:
            part = controller.model.add_partition(
                disk, size=1 << 20, flag="bios_grub")
        elif controller.model.bootloader == Bootloader.UEFI:
            part = controller.model.add_partition(
                disk, size=512 << 20, flag="boot")
        elif controller.model.bootloader == Bootloader.PREP:
            part = controller.model.add_partition(
                disk, size=8 << 20, flag="prep")
        part.preserve = True
        return part

    def assertIsBootDisk(self, controller, disk):
        if controller.model.bootloader == Bootloader.BIOS:
            self.assertTrue(disk.grub_device)
            self.assertEqual(disk.partitions()[0].flag, "bios_grub")
        elif controller.model.bootloader == Bootloader.UEFI:
            for part in disk.partitions():
                if part.flag == "boot" and part.grub_device:
                    return
            self.fail("{} is not a boot disk".format(disk))
        elif controller.model.bootloader == Bootloader.PREP:
            for part in disk.partitions():
                if part.flag == "prep" and part.grub_device:
                    self.assertEqual(part.wipe, 'zero')
                    return
            self.fail("{} is not a boot disk".format(disk))

    def assertIsNotBootDisk(self, controller, disk):
        if controller.model.bootloader == Bootloader.BIOS:
            self.assertFalse(disk.grub_device)
        elif controller.model.bootloader == Bootloader.UEFI:
            for part in disk.partitions():
                if part.flag == "boot" and part.grub_device:
                    self.fail("{} is a boot disk".format(disk))
        elif controller.model.bootloader == Bootloader.PREP:
            for part in disk.partitions():
                if part.flag == "prep" and part.grub_device:
                    self.fail("{} is a boot disk".format(disk))

    def test_boot_disk_resilient(self):
        for bl in Bootloader:
            if bl == Bootloader.NONE:
                continue
            controller = make_controller(bl)
            controller.supports_resilient_boot = True

            disk1 = make_disk(controller.model, preserve=False)
            disk2 = make_disk(controller.model, preserve=False)
            disk2p1 = controller.model.add_partition(
                disk2, size=disk2.free_for_partitions)

            controller.add_boot_disk(disk1)
            self.assertIsBootDisk(controller, disk1)
            if bl == Bootloader.UEFI:
                self.assertIsMountedAtBootEFI(disk1.partitions()[0])

            size_before = disk2p1.size
            controller.add_boot_disk(disk2)
            self.assertIsBootDisk(controller, disk1)
            self.assertIsBootDisk(controller, disk2)
            if bl == Bootloader.UEFI:
                self.assertIsMountedAtBootEFI(disk1.partitions()[0])
                self.assertNotMounted(disk2.partitions()[0])
            self.assertEqual(len(disk2.partitions()), 2)
            self.assertEqual(disk2.partitions()[1], disk2p1)
            self.assertEqual(
                disk2.partitions()[0].size + disk2p1.size, size_before)

            controller.remove_boot_disk(disk1)
            self.assertIsNotBootDisk(controller, disk1)
            self.assertIsBootDisk(controller, disk2)
            if bl == Bootloader.UEFI:
                self.assertIsMountedAtBootEFI(disk2.partitions()[0])
            self.assertEqual(len(disk1.partitions()), 0)

            controller.remove_boot_disk(disk2)
            self.assertIsNotBootDisk(controller, disk2)
            self.assertEqual(len(disk2.partitions()), 1)
            self.assertEqual(disk2p1.size, size_before)

    def test_boot_disk_no_resilient(self):
        for bl in Bootloader:
            if bl == Bootloader.NONE:
                continue
            controller = make_controller(bl)
            controller.supports_resilient_boot = False

            disk1 = make_disk(controller.model, preserve=False)
            disk2 = make_disk(controller.model, preserve=False)
            disk2p1 = controller.model.add_partition(
                disk2, size=disk2.free_for_partitions)

            controller.add_boot_disk(disk1)
            self.assertIsBootDisk(controller, disk1)
            if bl == Bootloader.UEFI:
                self.assertIsMountedAtBootEFI(disk1.partitions()[0])

            size_before = disk2p1.size
            controller.add_boot_disk(disk2)
            self.assertIsNotBootDisk(controller, disk1)
            self.assertIsBootDisk(controller, disk2)
            if bl == Bootloader.UEFI:
                self.assertIsMountedAtBootEFI(disk2.partitions()[0])
            self.assertEqual(len(disk2.partitions()), 2)
            self.assertEqual(disk2.partitions()[1], disk2p1)
            self.assertEqual(
                disk2.partitions()[0].size + disk2p1.size, size_before)

    def test_boot_disk_existing(self):
        for bl in Bootloader:
            if bl == Bootloader.NONE:
                continue
            controller = make_controller(bl)

            disk1 = make_disk(controller.model, preserve=True)
            part = self.add_existing_boot_partition(controller, disk1)

            wipe_before = part.wipe
            controller.add_boot_disk(disk1)
            self.assertIsBootDisk(controller, disk1)
            if bl == Bootloader.UEFI:
                self.assertIsMountedAtBootEFI(part)

            controller.remove_boot_disk(disk1)
            self.assertIsNotBootDisk(controller, disk1)
            self.assertEqual(len(disk1.partitions()), 1)
            self.assertEqual(part.wipe, wipe_before)
            if bl == Bootloader.UEFI:
                self.assertNotMounted(part)

    def test_mounting_partition_makes_boot_disk(self):
        controller = make_controller(Bootloader.UEFI)
        disk1 = make_disk(controller.model, preserve=True)
        disk1p1 = controller.model.add_partition(
            disk1, size=512 << 20, flag="boot")
        disk1p1.preserve = True
        disk1p2 = controller.model.add_partition(
            disk1, size=disk1.free_for_partitions)
        disk1p2.preserve = True
        controller.partition_disk_handler(
            disk1, disk1p2, {'fstype': 'ext4', 'mount': '/'})
        efi_mnt = controller.model._mount_for_path("/boot/efi")
        self.assertEqual(efi_mnt.device.volume, disk1p1)
